"""Cookbook internal module."""

import argparse
import importlib
import logging
import os
import sys
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import Optional, cast

from wmflib.config import load_yaml_config

from spicerack import Spicerack, SpicerackExtenderBase, _log, _module_api, cookbook
from spicerack._menu import BaseItem, CookbookItem, MenuError, TreeItem, get_module_title
from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


class CookbookError(SpicerackError):
    """Custom exception class for errors of this module."""


class CookbookCollection:
    """Collect and represent available cookbooks."""

    cookbooks_module_prefix: str = "cookbooks"

    def __init__(
        self,
        *,
        base_dirs: list[Path],
        args: Sequence[str],
        spicerack: Spicerack,
        path_filter: str = "",
    ) -> None:
        """Initialize the class and collect all the cookbook menu items.

        Arguments:
            base_dirs: the list of base directories from where to start looking for cookbooks.
            args: the list of arguments to pass to the collected items.
            spicerack: the initialized instance of the library.
            path_filter: an optional relative module path to filter for. If set, only cookbooks that are part of this
                subtree will be collected.

        """
        self.base_dirs = [base_dir / self.cookbooks_module_prefix for base_dir in base_dirs]
        self.args = args
        self.spicerack = spicerack
        self.path_filter = path_filter

        module = import_module(self.cookbooks_module_prefix)
        self.menu = TreeItem(module, self.args, self.spicerack, self.cookbooks_module_prefix)
        for base_dir in self.base_dirs:
            self._collect(base_dir)

    def get_item(self, path: str) -> Optional[BaseItem]:
        """Retrieve the item for a given path.

        Arguments:
            path: the path of the item to look for.

        Returns:
            None: when no item is found.
            spicerack._menu.TreeItem: when the item found is a menu of cookbooks.
            spicerack._menu.CookbookItem: when the item found is a single cookbook.

        """
        if path:
            effective_path = ".".join((self.cookbooks_module_prefix, path))
        else:
            effective_path = self.cookbooks_module_prefix

        item: Optional[BaseItem] = None
        parts = effective_path.split(".")
        multiple_class_name = ".".join(parts[-2:])
        for i, subpath in enumerate(parts):
            if i == 0:  # Initial menu
                item = self.menu
            elif item is not None and subpath in item.items.keys():  # Submenu or child cookbook found
                item = item.items[subpath]
            elif item is not None and i == len(parts) - 2 and multiple_class_name in item.items.keys():
                # Found cookbook in a module that has multiple class API cookbooks
                item = item.items[multiple_class_name]
                break
            else:
                item = None

        return item

    def _create_menu_for_path(self, path: str) -> TreeItem:
        """Create the menu for a given path, including intermediate levels, if missing. Return the existing one if any.

        Arguments:
            path: the path of the item to look for.

        """
        item: TreeItem = self.menu
        if path == self.cookbooks_module_prefix:
            return item

        progressive_path = [self.cookbooks_module_prefix]
        for subpath in path.split(".")[1:]:
            progressive_path.append(subpath)
            if subpath in item.items:
                item = cast(TreeItem, item.items[subpath])
            else:
                module_name = ".".join(progressive_path)
                submenu = TreeItem(
                    import_module(module_name),
                    self.args,
                    self.spicerack,
                    self.cookbooks_module_prefix,
                )
                # Use the parent owner_team if set
                default_owner = cookbook.CookbookBase.owner_team
                if item.owner_team != default_owner and submenu.owner_team == default_owner:
                    submenu.module.__owner_team__ = item.owner_team

                # When collecting the cookbooks and creating the TreeItem instances, the relation to the parent
                # menu should be skipped for those intermediate menus created for coherence but that should not be
                # accessible by the user, like when using a path_filter.
                add_parent = submenu.full_name.startswith(self.path_filter) and submenu.full_name != self.path_filter
                item.append(submenu, add_parent=add_parent)
                item = submenu

        return item

    def _collect(self, base_dir: Path) -> None:
        """Collect available cookbooks starting from a base path.

        Arguments:
            base_dir: the directory where to start collecting the cookbooks.

        """
        for dirpath in base_dir.rglob(""):  # Selects only directories
            if dirpath.name == "__pycache__":
                continue

            relpath = dirpath.relative_to(base_dir)
            if relpath.name:
                prefix = str(relpath).replace("/", ".").rstrip(".")
                module_prefix = f"{self.cookbooks_module_prefix}.{prefix}"
            else:
                prefix = ""
                module_prefix = self.cookbooks_module_prefix

            if self._should_filter(prefix):
                continue

            try:
                menu = self._create_menu_for_path(module_prefix)
            except CookbookError as e:
                logger.error(e)
                continue

            for filepath in dirpath.glob("[!_]*.py"):  # Excludes files starting with an underscore, like __init__.py
                module_name = f"{module_prefix}.{filepath.stem}"
                self._collect_filename(module_name, menu)

    def _collect_filename(self, module_name: str, menu: TreeItem) -> None:
        """Collect all the available cookbooks in the given module and add them to the menu.

        Arguments:
            module_name: the Python module to load.
            menu: the menu to append the collected cookbook to.

        """
        try:
            classes = self._collect_module_cookbooks(import_module(module_name))
        except CookbookError as e:
            logger.error(e)
            return

        for class_obj in classes:
            try:
                cookbook_item = CookbookItem(class_obj, self.args, self.spicerack)
            except MenuError as e:
                logger.error(e)
                continue

            # Use the parent owner_team if set
            default_owner = cookbook.CookbookBase.owner_team
            if menu.owner_team != default_owner and cookbook_item.owner_team == default_owner:
                cookbook_item.instance.owner_team = menu.owner_team

            if self._should_filter(cookbook_item.full_name):
                continue

            menu.append(cookbook_item)

    def _should_filter(self, name: str) -> bool:
        """Check if a given path or full name should be skipped because not matching the current filter.

        Arguments:
            name: the name to check, can be either a CookbookItem full name or a TreeItem path.

        Returns:
            :py:data:`True` if the item should be skipped, :py:data:`False` otherwise.

        """
        if not self.path_filter:
            return False

        if name.startswith(self.path_filter[: len(name)]):
            return False

        return True

    def _collect_module_cookbooks(
        self, module: _module_api.CookbooksModuleInterface
    ) -> list[type[cookbook.CookbookBase]]:
        """Collect all classes derived from CookbookBase in the given module.

        Arguments:
            module: the module to check for cookbook classes.

        Returns:
            list: a list of CookbookBase derived class objects.

        """
        attrs = [getattr(module, attr) for attr in dir(module)]
        classes = [
            attr
            for attr in attrs
            if isinstance(attr, type) and issubclass(attr, cookbook.CookbookBase) and attr.__module__ == module.__name__
        ]

        if not classes:  # No class API cookbook found, convert a module API cookbook into a class
            classes.append(self._convert_module_in_cookbook(module))
        elif len(classes) == 1:  # Avoid the unnecessary further nested namespace
            class_obj = classes[0]
            full_name = class_obj.__module__.split(".", 1)[1]
            if "." in full_name:
                class_obj.spicerack_path, class_obj.spicerack_name = full_name.rsplit(".", 1)
            else:
                class_obj.spicerack_path = ""
                class_obj.spicerack_name = full_name
        else:  # Inject the class module name as part of the class name and set the module as that of the parent
            for class_obj in classes:
                full_name_prefix = class_obj.__module__.split(".", 1)[1]
                if "." in full_name_prefix:
                    class_obj.spicerack_path, name_prefix = full_name_prefix.rsplit(".", 1)
                    class_obj.spicerack_name = ".".join([name_prefix, class_obj.__name__])
                else:
                    class_obj.spicerack_path = ""
                    class_obj.spicerack_name = ".".join([full_name_prefix, class_obj.__name__])

        return classes

    def _convert_module_in_cookbook(  # noqa: MC0001
        self, module: _module_api.CookbooksModuleInterface
    ) -> type[cookbook.CookbookBase]:
        """Convert a module API based cookbook into a class API cookbook.

        Arguments:
            module: the module to convert.

        Returns:
            type: a dynamically generated class derived from :py:class:`spicerack.cookbook.CookbookBase`.

        """
        module_name, name = module.__name__.rsplit(".", 1)
        title = get_module_title(module) or CookbookItem.fallback_title

        try:
            owner_team = module.__owner_team__
        except AttributeError:
            owner_team = cookbook.CookbookBase.owner_team

        try:
            run = module.run
        except AttributeError as e:
            raise CookbookError(f"Unable to find run function in module {module.__name__}") from e

        if hasattr(module, "MAX_CONCURRENCY"):
            max_concurrency = module.MAX_CONCURRENCY
        else:
            max_concurrency = cookbook.CookbookRunnerBase.max_concurrency

        if hasattr(module, "LOCK_TTL"):
            lock_ttl = module.LOCK_TTL
        else:
            lock_ttl = cookbook.CookbookRunnerBase.lock_ttl

        runner_name = f"{name}Runner"
        runner = type(
            runner_name,
            (_module_api.CookbookModuleRunnerBase,),
            {
                "_run": staticmethod(run),
                "max_concurrency": max_concurrency,
                "lock_ttl": lock_ttl,
            },
        )
        runner.__module__ = module_name

        attributes = {
            "__module__": module_name,
            "__name__": name,
            "owner_team": owner_team,
            "spicerack_name": name,
            "spicerack_path": module_name.split(".", 1)[1] if "." in module_name else "",
            "title": title,
            "get_runner": lambda _, args: runner(args, self.spicerack),
        }

        try:
            args_parser = module.argument_parser
            attributes["argument_parser"] = lambda _: args_parser()
        except AttributeError:
            # The cookbook doesn't accept any argument, set a default empty parser
            attributes["argument_parser"] = lambda _: argparse.ArgumentParser(
                description=module.__doc__, formatter_class=cookbook.ArgparseFormatter
            )

        cookbook_class = type(name, (cookbook.CookbookBase,), attributes)
        return cookbook_class


def argument_parser() -> argparse.ArgumentParser:
    """Get the CLI argument parser.

    If the COOKBOOK is passed as a path, it will be converted to a Python module syntax.

    """
    parser = argparse.ArgumentParser(
        description="Spicerack Cookbook Runner",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="See also https://wikitech.wikimedia.org/wiki/Spicerack/Cookbooks",
        allow_abbrev=False,  # Prevent matching of cookbook options
    )
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help=(
            "List all available cookbooks and their owner team, if -v/--verbose is set print also their description. "
            "If a COOKBOOK is also specified, it will be used as a prefix filter."
        ),
    )
    parser.add_argument(
        "-c",
        "--config-file",
        default="/etc/spicerack/config.yaml",
        help="Path to the Spicerack configuration file to load.",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Set the DRY-RUN mode, also for the cookbook.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose output, also for the cookbook.",
    )
    parser.add_argument(
        "--no-locks",
        action="store_true",
        help="Do not acquire or check distributed locks on etcd. To be used only if there are issues with etcd.",
    )
    parser.add_argument(
        "cookbook",
        metavar="COOKBOOK",
        nargs="?",
        type=cookbook_path_type,
        default="",
        help=(
            "Either the name of the Python module to execute (sre.hosts.downtime) or the relative path of the Python "
            "file to execute (sre/hosts/downtime.py).  If the selected path/module is a directory (sre.hosts) or is "
            "not set, an interactive menu will be shown. Each directory listed in the `cookbooks_base_dirs` key of the"
            "configuration file represents a checked out cookbook repository containing a `cookbooks` subfolder. All "
            "paths specified must be relative to theses subfolders."
        ),
    )
    parser.add_argument(
        "cookbook_args",
        metavar="COOKBOOK_ARGS",
        nargs=argparse.REMAINDER,
        help="Collect all the remaining arguments to be passed to the cookbook or menu to execute.",
    )

    return parser


def cookbook_path_type(path: str) -> str:
    """Convert a COOKBOOK path to module syntax, if it's in path syntax.

    Arguments:
        path: the path to be converted.

    """
    if path.endswith(".py"):
        path = path[:-3].replace("/", ".")

    return path


def import_module(module_name: str) -> _module_api.CookbooksModuleInterface:
    """Import a Python module and return it.

    Arguments:
        module_name: the name of the module to load.

    Raises:
        spicerack._cookbook.CookbookError: on failure to load the module.

    """
    try:
        module = importlib.import_module(module_name)
    except Exception as e:
        raise CookbookError(f"Failed to import module {module_name}: {e}") from e

    return cast(_module_api.CookbooksModuleInterface, module)


def _import_extender_class(extender_class_param: str) -> SpicerackExtenderBase:
    """Import the extender class, ensure it's a subclass of SpicerackExtenderBase and return it.

    Arguments:
        extender_class_param: the configuration file parameter with the fully qualified class name.

    """
    extender_module_name, extender_class_name = extender_class_param.rsplit(".", 1)
    extender_module = import_module(extender_module_name)
    extender_class = getattr(extender_module, extender_class_name)
    if issubclass(extender_class, SpicerackExtenderBase):
        return extender_class

    raise TypeError(f"Extender class {extender_class_param} is not a subclass of spicerack.SpicerackExtenderBase.")


def get_cookbook_callback(
    cookbooks_base_dirs: list[Path],
) -> Callable[["Spicerack", str, Sequence[str]], Optional["BaseItem"]]:
    """Returns the cookbook callback function needed to load a cookbook from another cookbook."""

    def get_cookbook(spicerack: Spicerack, cookbook_path: str, cookbook_args: Sequence[str] = ()) -> Optional[BaseItem]:
        """Get a cookbook item if it exists.

        Arguments:
            spicerack: the Spicerack class instance.
            cookbook_path: the cookbook name/path.
            cookbook_args: the sequence of CLI arguments to pass to the cookbook.

        Returns:
            :py:data:`None` if there is no cookbook found, the cookbook item otherwise.

        """
        cookbooks = CookbookCollection(
            base_dirs=cookbooks_base_dirs,
            args=cookbook_args,
            spicerack=spicerack,
            path_filter=cookbook_path,
        )
        return cookbooks.get_item(cookbook_path)

    return get_cookbook


def main(argv: Optional[Sequence[str]] = None) -> Optional[int]:  # noqa: MC0001
    """Entry point, run the tool.

    Arguments:
        argv: the list of command line arguments to parse.

    Returns:
        The return code, zero on success, non-zero on failure.

    """
    args = argument_parser().parse_args(argv)
    if not args.cookbook:
        args.cookbook = ""

    config = load_yaml_config(args.config_file)
    cookbooks_base_dirs = []
    for base_dir in config["cookbooks_base_dirs"]:
        base_dir_path = Path(base_dir).expanduser()
        cookbooks_base_dirs.append(base_dir_path)
        sys.path.append(str(base_dir_path))

    if not cookbooks_base_dirs:
        print(
            "No cookbooks paths are specified in the `cookbooks_base_dirs` key of the configuration file.",
            file=sys.stderr,
        )
        return 1

    if config.get("external_modules_dir") is not None:
        sys.path.append(str(Path(config["external_modules_dir"]).expanduser()))

    params = config.get("instance_params", {})
    get_cookbook = get_cookbook_callback(cookbooks_base_dirs)
    params.update({"verbose": args.verbose, "dry_run": args.dry_run, "get_cookbook_callback": get_cookbook})
    if "extender_class" in params:
        try:
            params["extender_class"] = _import_extender_class(params["extender_class"])
        except Exception as e:  # pylint: disable=broad-except
            print(f"Failed to import the extender_class {params['extender_class']}:", e, file=sys.stderr)
            return 1

    if args.no_locks:
        params["etcd_config"] = ""  # Disable locking support

    try:
        spicerack = Spicerack(**params)
    except TypeError as e:
        print("Unable to instantiate Spicerack, check your configuration:", e, file=sys.stderr)
        return 1

    cookbooks = CookbookCollection(
        base_dirs=cookbooks_base_dirs, args=args.cookbook_args, spicerack=spicerack, path_filter=args.cookbook
    )
    if args.list:
        print(cookbooks.menu.get_tree(), end="")
        return 0

    cookbook_item = cookbooks.get_item(args.cookbook)
    if cookbook_item is None:
        logger.error("Unable to find cookbook %s", args.cookbook)
        return cookbook.NOT_FOUND_RETCODE

    base_path = Path(config["logs_base_dir"]).expanduser() / cookbook_item.path.replace(".", os.sep)
    _log.setup_logging(
        base_path,
        cookbook_item.name,
        spicerack.username,
        dry_run=args.dry_run,
        host=config.get("tcpircbot_host", None),
        port=int(config.get("tcpircbot_port", 0)),
        notify_logger_enabled=config.get("user_input_notifications_enabled", False),
    )

    logger.debug("Executing cookbook %s with args: %s", args.cookbook, args.cookbook_args)
    return cookbook_item.run()



"""Orchestrator API module."""

import logging
from typing import Any, Type, Union

from requests import Response

try:  # JSONDecodeError present since requests 2.27.0, bullseye has 2.25.1
    from requests.exceptions import JSONDecodeError

    CompatJSONDecodeError: Union[Type[ValueError], Type[JSONDecodeError]] = JSONDecodeError
except ImportError:
    CompatJSONDecodeError = ValueError

from spicerack.apiclient import APIClient, APIClientResponseError
from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)
DEFAULT_PORT = 3306


class OrchestratorError(SpicerackError):
    """General errors raised by this module."""


class Orchestrator(APIClient):
    """Orchestrator API class."""

    def request(self, *args: Any, **kwargs: Any) -> Response:
        """Perform an API request to Orchestrator based on the given parameters.

        See :py:meth:`spicerack.apiclient.APIClient.request` for all the arguments and return values.

        Raises:
            spicerack.apiclient.APIClientError: if the given uri does not start with a slash (/) or if the request
                couldn't be performed.
            spicerack.orchestrator.OrchestratorError: if the response status code is not ok, between 400 and 600.
                It will parse the JSON response and use it in the exception message if present.

        """
        try:
            return super().request(*args, **kwargs)
        except APIClientResponseError as e:
            try:  # Orchestrator tends to response 500 with a JSON on some errors
                response = e.response.json()
                message = f"[{response['Code']}] {response['Message']}"
                if response["Details"]:
                    message += f" {response['Details']}"
            except CompatJSONDecodeError:  # but also responds text on other errors like 404
                message = f"{e.response.text}"

            raise OrchestratorError(message) from e

    def clusters(self) -> dict[str, dict]:
        """Get the current clusters and their metadata.

        Returns:
            a dictionary of clusters with cluster names as keys and cluster metadata as values.

        Raises:
            spicerack.apiclient.APIClient: if the request fails for any reason.

        """
        return {cluster["ClusterAlias"]: cluster for cluster in self.request("get", "/clusters-info").json()}

    def instance(self, instance: str) -> dict:
        """Get the metadata of an instance. It will automatically splits the port if present or use the default if not.

        Arguments:
            instance: the instance name as reported in orchestrator (``FQDN:PORT``). If the port is not present it will
                automatically use the :py:const:`spicerack.orchestrator.DEFAULT_PORT`.

        Returns:
            a dictionary with the instance metadata.

        Raises:
            spicerack.apiclient.APIClient: if the request fails for any reason or the instance does not exists.

        """
        host, port = instance.rsplit(":", 1) if ":" in instance else (instance, str(DEFAULT_PORT))
        return self.request("get", f"/instance/{host}/{port}").json()



"""Orchestrator module tests."""

import re

import pytest
from wmflib.requests import http_session

from spicerack import orchestrator


class TestOrchestrator:
    """Tests for the Orchestrator class."""

    @pytest.fixture(autouse=True)
    def setup_method(self, requests_mock):
        """Initialize the test instance."""
        # pylint: disable=attribute-defined-outside-init
        self.requests_mock = requests_mock
        self.base_url = "https://orchestrator.example.org/api"
        self.endpoint = f"{self.base_url}/endpoint"
        self.http_session = http_session("Orchestrator Tests")
        self.http_session.headers.update({"Accept": "application/json"})
        self.orchestrator = orchestrator.Orchestrator(self.base_url, self.http_session, dry_run=False)

    def test_request_ok(self):
        """It should perform the provided request and return it."""
        self.requests_mock.get(self.endpoint, json={"key": "value"})
        response = self.orchestrator.request("get", "/endpoint")
        assert response.json() == {"key": "value"}
        assert response.status_code == 200

    @pytest.mark.parametrize("details", (None, "Additional details"))
    def test_request_response_wrong_status_code(self, details):
        """It should raise an OrchestratorError if the request returns an error status code parsing the message."""
        response = {"Code": "ERROR", "Message": "Cannot read instance: db.example.org:3307", "Details": details}
        self.requests_mock.get(self.endpoint, json=response, status_code=500)
        expected = "[ERROR] Cannot read instance: db.example.org:3307"
        if details:
            expected += f" {details}"
        with pytest.raises(orchestrator.OrchestratorError, match=re.escape(expected)):
            self.orchestrator.request("get", "/endpoint")

    def test_request_response_wrong_status_code_failed_json(self):
        """It should raise an OrchestratorError if the request fails and it's unable to parse the message."""
        self.requests_mock.get(self.endpoint, text="404 page not found", status_code=404)
        with pytest.raises(orchestrator.OrchestratorError, match="404 page not found"):
            self.orchestrator.request("get", "/endpoint")

    def test_clusters(self):
        """It should return the clusters known to orchestrator with their metadata."""
        response = [{"ClusterAlias": "section1", "other": "info1"}, {"ClusterAlias": "section2", "other": "info2"}]
        self.requests_mock.get(self.base_url + "/clusters-info", json=response)
        clusters = self.orchestrator.clusters()
        assert list(clusters.keys()) == ["section1", "section2"]
        assert clusters["section2"]["other"] == "info2"

    def test_instance_default_port(self):
        """It should return the instance metadata appending the default port."""
        response = {"Key": {"Hostname": "db.example.org", "Port": 3306}, "Uptime": 123456}
        self.requests_mock.get(f"{self.base_url}/instance/db.example.org/3306", json=response)
        instance = self.orchestrator.instance("db.example.org")
        assert instance["Uptime"] == 123456

    def test_instance_custom_port(self):
        """It should return the instance metadata."""
        response = {"Key": {"Hostname": "db.example.org", "Port": 3333}, "Uptime": 123456}
        self.requests_mock.get(f"{self.base_url}/instance/db.example.org/3333", json=response)
        instance = self.orchestrator.instance("db.example.org:3333")
        assert instance["Uptime"] == 123456



# Group2 zCookbook4. No docstring to test the fallback value.
# pylint: disable=missing-module-docstring
# noqa:  D100


def run(_args, _spicerack):
    """As required by spicerack._cookbook."""
    return 0



"""Dbctl module tests."""

from unittest import mock

import pytest
from conftool import configuration, kvobject
from conftool.tests.unit import MockBackend

from spicerack import dbctl
from spicerack.tests import get_fixture_path


class TestDbctl:
    """Dbctl test class."""

    def setup_method(self):
        """Setup a Dbctl instance with a mocked conftool backend and driver."""
        # pylint: disable=attribute-defined-outside-init
        self.conftool_backend = MockBackend({})
        kvobject.KVObject.backend = self.conftool_backend
        kvobject.KVObject.config = configuration.Config(driver="")
        self.config = get_fixture_path("confctl", "config.yaml")
        self.schema = get_fixture_path("confctl", "schema.yaml")
        with mock.patch("spicerack.confctl.kvobject.KVObject.setup"):
            self.dbctl = dbctl.Dbctl(config=self.config, schema=self.schema, dry_run=False)

    @pytest.mark.parametrize(
        "property_name, instance",
        (
            ("config", dbctl.DbConfig),
            ("instance", dbctl.Instance),
            ("section", dbctl.Section),
        ),
    )
    def test_init(self, property_name, instance):
        """It should give access to the dbconfig instances for the various type of objects."""
        assert isinstance(getattr(self.dbctl, property_name), instance)



"""Cookbook module tests."""

import logging
import shutil
from pathlib import Path
from unittest import mock

import pytest

from spicerack import Spicerack, _cookbook, _menu, cookbook
from spicerack.tests import SPICERACK_TEST_PARAMS, get_fixture_path
from spicerack.tests.unit.test__log import reset_logging_module

COOKBOOKS_BASE_PATH = Path("spicerack/tests/fixtures/cookbook")
COOKBOOKS_BASE_PATHS = [COOKBOOKS_BASE_PATH]
COOKBOOKS_BASE_PATHS_MULTI = COOKBOOKS_BASE_PATHS + [Path("spicerack/tests/fixtures/external_cookbook")]
LIST_COOKBOOKS_ALL = """cookbooks
|-- class_api
|   |-- class_api.call_another_cookbook [team1]
|   |-- class_api.example [unowned]
|   |-- class_api.get_runner_raise [unowned]
|   |-- class_api.init_abort [team1]
|   |-- class_api.lock_args [unowned]
|   |-- class_api.multiple.CookbookA [team1]
|   |-- class_api.multiple.CookbookB [team2]
|   |-- class_api.rollback [unowned]
|   |-- class_api.rollback_raise [unowned]
|   |-- class_api.runtime_description [unowned]
|   |-- class_api.runtime_description_raise [unowned]
|   `-- class_api.use_external_modules [unowned]
|-- cookbook [unowned]{external_cookbooks}
|-- group1
|   `-- group1.cookbook1 [unowned]
|-- group2
|   |-- group2.cookbook2 [team2]
|   |-- group2.subgroup1
|   |   `-- group2.subgroup1.cookbook3 [team9]
|   `-- group2.zcookbook4 [team2]
|-- group3
|   |-- group3.argparse_ok [team1]
|   |-- group3.argument_parser_raise_system_exit [team3]
|   |-- group3.get_argument_parser_raise [team3]
|   |-- group3.get_argument_parser_raise_system_exit_str [team3]
|   |-- group3.keyboard_interrupt [team3]
|   |-- group3.non_zero_exit [team3]
|   |-- group3.raise_exception [team3]
|   |-- group3.raise_system_exit_0 [team3]
|   |-- group3.raise_system_exit_9 [team3]
|   |-- group3.raise_system_exit_str [team3]
|   `-- group3.subgroup3
|       `-- group3.subgroup3.cookbook4 [team9]
|-- multiple.CookbookA [team1]
|-- multiple.CookbookB [team2]
`-- root [unowned]
"""
LIST_EXTERNAL_COOKBOOKS = """
|-- external_cookbook [unowned]
|-- external_group
|   `-- external_group.cookbook1 [unowned]"""
LIST_COOKBOOKS_ALL_VERBOSE = """cookbooks
|-- class_api: Class API Test Cookbooks.
|   |-- class_api.call_another_cookbook: A cookbook that calls another cookbook. [team1]
|   |-- class_api.example: - [unowned]
|   |-- class_api.get_runner_raise: Class API get_runner raise cookbook. [unowned]
|   |-- class_api.init_abort: A cookbook that aborts successfully in its runner's init. [team1]
|   |-- class_api.lock_args: Cookbook that overrides the lock arguments. [unowned]
|   |-- class_api.multiple.CookbookA: Multiple cookbook classes. [team1]
|   |-- class_api.multiple.CookbookB: Multiple cookbook classes. [team2]
|   |-- class_api.rollback: Class API rollback cookbook. [unowned]
|   |-- class_api.rollback_raise: Class API rollback_raise cookbook. [unowned]
|   |-- class_api.runtime_description: Class API cookbook that overrides runtime_description. [unowned]
|   |-- class_api.runtime_description_raise: Class API runtime_description raise cookbook. [unowned]
|   `-- class_api.use_external_modules: Call a Spicerack extender accessor. [unowned]
|-- cookbook: Top level class cookbook. [unowned]
|-- group1: Group1 Test Cookbooks.
|   `-- group1.cookbook1: Group1 Cookbook1. [unowned]
|-- group2: -
|   |-- group2.cookbook2: Group2 Cookbook2. [team2]
|   |-- group2.subgroup1: Group2 Subgroup1 Test Cookbooks.
|   |   `-- group2.subgroup1.cookbook3: Group2 Subgroup1 Cookbook3. [team9]
|   `-- group2.zcookbook4: UNKNOWN (unable to detect title) [team2]
|-- group3: Group3 Test Cookbooks.
|   |-- group3.argparse_ok: Group3 argparse_ok. [team1]
|   |-- group3.argument_parser_raise_system_exit: Group3 argument_parser() raise SystemExit. [team3]
|   |-- group3.get_argument_parser_raise: Group3 get argument_parser() raise. [team3]
|   |-- group3.get_argument_parser_raise_system_exit_str: Group3 get argument_parser() raise SystemExit(str). [team3]
|   |-- group3.keyboard_interrupt: Group3 Raise KeyboardInterrupt. [team3]
|   |-- group3.non_zero_exit: Group3 Non-Zero return code. [team3]
|   |-- group3.raise_exception: Group3 Raise Exception. [team3]
|   |-- group3.raise_system_exit_0: Group3 Raise SystemExit(0). [team3]
|   |-- group3.raise_system_exit_9: Group3 Raise SystemExit(9). [team3]
|   |-- group3.raise_system_exit_str: Group3 Raise SystemExit('message'). [team3]
|   `-- group3.subgroup3: Group3 Subgroup3 Test Cookbooks.
|       `-- group3.subgroup3.cookbook4: Group3 Subgroup3 Cookbook4. [team9]
|-- multiple.CookbookA: Multiple cookbook classes. [team1]
|-- multiple.CookbookB: Multiple cookbook classes. [team2]
`-- root: Top level cookbook. [unowned]
"""
LIST_COOKBOOKS_GROUP3 = """cookbooks
`-- group3
    |-- group3.argparse_ok [team1]
    |-- group3.argument_parser_raise_system_exit [team3]
    |-- group3.get_argument_parser_raise [team3]
    |-- group3.get_argument_parser_raise_system_exit_str [team3]
    |-- group3.keyboard_interrupt [team3]
    |-- group3.non_zero_exit [team3]
    |-- group3.raise_exception [team3]
    |-- group3.raise_system_exit_0 [team3]
    |-- group3.raise_system_exit_9 [team3]
    |-- group3.raise_system_exit_str [team3]
    `-- group3.subgroup3
        `-- group3.subgroup3.cookbook4 [team9]
"""
LIST_COOKBOOKS_GROUP3_SUBGROUP3 = """cookbooks
`-- group3
    `-- group3.subgroup3
        `-- group3.subgroup3.cookbook4 [team9]
"""
COOKBOOKS_MENU_TTY = """#--- cookbooks args=[] ---#
[0/12] class_api: Class API Test Cookbooks.
[NOTRUN] cookbook: Top level class cookbook.
[0/1] group1: Group1 Test Cookbooks.
[0/3] group2: -
[0/11] group3: Group3 Test Cookbooks.
[NOTRUN] multiple.CookbookA: Multiple cookbook classes.
[NOTRUN] multiple.CookbookB: Multiple cookbook classes.
[NOTRUN] root: Top level cookbook.
q - Quit
h - Help
"""
COOKBOOKS_MENU_NOTTY = """#--- cookbooks args=[] ---#
[0/12] class_api: Class API Test Cookbooks.
[NOTRUN] cookbook: Top level class cookbook.
[0/1] group1: Group1 Test Cookbooks.
[0/3] group2: -
[0/11] group3: Group3 Test Cookbooks.
[NOTRUN] multiple.CookbookA: Multiple cookbook classes.
[NOTRUN] multiple.CookbookB: Multiple cookbook classes.
[NOTRUN] root: Top level cookbook.
q - Quit
h - Help
Not a tty, exiting.
"""
COOKBOOKS_GROUP1_MENU = """#--- Group1 Test Cookbooks. args=[] ---#
[NOTRUN] cookbook1: Group1 Cookbook1.
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_MENU = """#--- group2 args=[] ---#
[NOTRUN] cookbook2: Group2 Cookbook2.
[0/1] subgroup1: Group2 Subgroup1 Test Cookbooks.
[NOTRUN] zcookbook4: UNKNOWN (unable to detect title)
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN = """[{'argument': None, 'k': False}, False, False]
#--- group2 args=[] ---#
[PASS] cookbook2: Group2 Cookbook2.
[0/1] subgroup1: Group2 Subgroup1 Test Cookbooks.
[NOTRUN] zcookbook4: UNKNOWN (unable to detect title)
b - Back to parent menu
h - Help
"""
COOKBOOKS_GROUP2_SUBGROUP1_MENU = """#--- Group2 Subgroup1 Test Cookbooks. args=[] ---#
[NOTRUN] cookbook3: Group2 Subgroup1 Cookbook3.
b - Back to parent menu
h - Help
"""


def test_argument_parser_converts_path():
    """It should convert a path-based cookbook into a module path."""
    argv = ["group1/cookbook1.py"]
    args = _cookbook.argument_parser().parse_args(argv)
    assert args.cookbook == "group1.cookbook1"


def test_argument_parser_keeps_module():
    """It should keep a module-based cookbook as it is."""
    argv = ["group1.cookbook1"]
    args = _cookbook.argument_parser().parse_args(argv)
    assert args.cookbook == argv[0]


def test_argument_parser_accept_empty():
    """With no args it should not fail."""
    args = _cookbook.argument_parser().parse_args([])
    assert args.cookbook == ""


def test_parse_args_list():
    """Passing -l/--list should keep the cookbook None and set list."""
    argv = ["--list"]
    args = _cookbook.argument_parser().parse_args(argv)
    assert args.list
    assert args.cookbook == ""


class TestMain:
    """Test class for the main function."""

    def teardown_method(self):
        """Tear down the test setting resetting the logging module."""
        reset_logging_module()

    def test_main_wrong_instance_config(self, capsys):
        """If the configuration file has invalid instance_params it should print an error and exit."""
        ret = _cookbook.main(["-c", str(get_fixture_path("config_wrong_overrides.yaml")), "cookbook"])
        _, err = capsys.readouterr()
        assert ret == 1
        assert "Unable to instantiate Spicerack, check your configuration" in err

    @pytest.mark.parametrize("confirm", (True, False))
    @pytest.mark.parametrize("raises", (True, False))
    def test_main_call_another_cookbook_ok(self, raises, confirm, capsys, caplog):
        """It should execute the cookbook that calls another cookbook."""
        args = ["-c", str(get_fixture_path("config.yaml")), "class_api.call_another_cookbook", "class_api.example"]
        if raises:
            args.append("--raises")
        if confirm:
            args.append("--confirm")

        with caplog.at_level(logging.DEBUG):
            ret = _cookbook.main(args)

        _, err = capsys.readouterr()
        assert ret == 0
        expected = [
            "START - Cookbook class_api.call_another_cookbook",
            "START - Cookbook class_api.example",
            "END (PASS) - Cookbook class_api.example (exit_code=0)",
            "END (PASS) - Cookbook class_api.call_another_cookbook (exit_code=0)",
        ]
        for line in expected:
            assert line in err

        assert "__COOKBOOK_STATS__:name=class_api.call_another_cookbook,exit_code=0,duration=" in caplog.text
        assert "__COOKBOOK_STATS__:name=class_api.example,exit_code=0,duration=" in caplog.text

    def test_main_call_another_cookbook_not_found(self, capsys):
        """It should fail to call another cookbook if it doesn't exists."""
        ret = _cookbook.main(
            ["-c", str(get_fixture_path("config.yaml")), "class_api.call_another_cookbook", "class_api.not_existent"]
        )
        _, err = capsys.readouterr()
        assert ret == cookbook.EXCEPTION_RETCODE
        assert "SpicerackError: Unable to find cookbook class_api.not_existent" in err
        assert "END (FAIL) - Cookbook class_api.call_another_cookbook (exit_code=99)" in err

    def test_main_use_external_modules_and_extender_ok(self, capsys):
        """It should inject the external module into the path and execute the cookbook with the extender."""
        ret = _cookbook.main(
            ["-c", str(get_fixture_path("config_external_modules.yaml")), "class_api.use_external_modules"]
        )
        _, err = capsys.readouterr()
        assert ret == 0
        expected = [
            "START - Cookbook class_api.use_external_modules",
            "Extender is a cool feature!",
            "END (PASS) - Cookbook class_api.use_external_modules (exit_code=0)",
        ]
        for line in expected:
            assert line in err

    def test_main_use_external_modules_and_extender_raise(self, capsys):
        """It should exit with 1 if unable to load the extender class."""
        ret = _cookbook.main(
            ["-c", str(get_fixture_path("config_bad_external_modules.yaml")), "class_api.use_external_modules"]
        )
        _, err = capsys.readouterr()
        assert ret == 1
        assert "Failed to import the extender_class spicerack_extender.SpicerackBadExtender" in err

    def test_main_empty_cookbooks_base_dirs_raise(self, capsys):
        """It should exit with 1 if there are no cookbooks paths specified in cookbooks_base_dirs."""
        ret = _cookbook.main(["-c", str(get_fixture_path("config_empty_base_dirs.yaml")), "-l"])
        _, err = capsys.readouterr()
        assert ret == 1
        assert "No cookbooks paths are specified in the `cookbooks_base_dirs` key of the configuration file." in err


class TestCookbookCollection:
    """Test class for the CookbookCollection class."""

    def setup_method(self):
        """Initialize the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.spicerack = Spicerack(verbose=False, dry_run=False, **SPICERACK_TEST_PARAMS)
        self.spicerack_verbose = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    def teardown_method(self):
        """Tear down the test setting resetting the logging module."""
        reset_logging_module()

    @pytest.mark.parametrize(
        "path_filter, verbose, expected",
        (
            ("", False, LIST_COOKBOOKS_ALL.format(external_cookbooks="")),
            ("", True, LIST_COOKBOOKS_ALL_VERBOSE),
            ("group3", False, LIST_COOKBOOKS_GROUP3),
            ("group3.subgroup3", False, LIST_COOKBOOKS_GROUP3_SUBGROUP3),
            ("group3.non_existent", False, ""),
        ),
    )
    def test_cookbooks_str(self, monkeypatch, path_filter, verbose, expected):
        """The string representation of the CookbookCollection should print all the cookbooks as a tree."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        if not path_filter and not verbose:  # For the first test ensure there is no __pycache__ directory.
            shutil.rmtree(COOKBOOKS_BASE_PATH / "cookbooks" / "__pycache__", ignore_errors=True)

        spicerack = self.spicerack
        if verbose:
            spicerack = self.spicerack_verbose
        cookbooks = _cookbook.CookbookCollection(
            base_dirs=COOKBOOKS_BASE_PATHS, args=[], spicerack=spicerack, path_filter=path_filter
        )
        assert cookbooks.menu.get_tree() == expected

    def test_cookbooks_multiple_dirs_str(self, monkeypatch):
        """The string representation of the CookbookCollection should print all the cookbooks from multiple dirs."""
        for base_path in COOKBOOKS_BASE_PATHS_MULTI:
            monkeypatch.syspath_prepend(base_path)
            shutil.rmtree(base_path / "cookbooks" / "__pycache__", ignore_errors=True)

        spicerack = self.spicerack
        cookbooks = _cookbook.CookbookCollection(base_dirs=COOKBOOKS_BASE_PATHS_MULTI, args=[], spicerack=spicerack)
        assert cookbooks.menu.get_tree() == LIST_COOKBOOKS_ALL.format(external_cookbooks=LIST_EXTERNAL_COOKBOOKS)

    def test_cookbooks_non_existent(self):
        """The CookbookCollection class initialized with an empty path should not collect any cookbook."""
        cookbooks = _cookbook.CookbookCollection(
            base_dirs=[COOKBOOKS_BASE_PATH / "non_existent"], args=[], spicerack=self.spicerack
        )
        assert cookbooks.menu.get_tree() == ""

    @pytest.mark.parametrize(
        "module, messages, absent_messages, code, args",
        (
            (
                "cookbook",
                [
                    "START - Cookbook cookbook",
                    "END (PASS) - Cookbook cookbook (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            ("cookbook", [], ["START - Cookbook", "END ("], 2, ["arg1"]),
            (
                "group3.non_zero_exit",
                ["END (FAIL) - Cookbook group3.non_zero_exit (exit_code=1)"],
                [],
                1,
                [],
            ),
            (
                "group3.non_zero_exit",
                ["END (FAIL) - Cookbook group3.non_zero_exit (exit_code=1)"],
                [],
                1,
                [],
            ),
            (
                "group3.non_existent",
                ["Unable to find cookbook"],
                [],
                cookbook.NOT_FOUND_RETCODE,
                [],
            ),
            ("group3.argparse_ok", [], ["START - Cookbook", "END ("], 0, ["-h"]),
            (
                "group3.invalid_syntax",
                ["invalid syntax (invalid_syntax.py, line 6)"],
                [],
                cookbook.NOT_FOUND_RETCODE,
                [],
            ),
            (
                "group3.keyboard_interrupt",
                ["Ctrl+c pressed"],
                [],
                cookbook.INTERRUPTED_RETCODE,
                [],
            ),
            (
                "group3.get_argument_parser_raise",
                ["raised while getting argument parser for cookbook"],
                ["START - Cookbook", "END ("],
                cookbook.GET_ARGS_PARSER_FAIL_RETCODE,
                [],
            ),
            (
                "group3.get_argument_parser_raise_system_exit_str",
                ["SystemExit('argument_parser')"],
                ["START - Cookbook", "END ("],
                cookbook.GET_ARGS_PARSER_FAIL_RETCODE,
                [],
            ),
            (
                "group3.argument_parser_raise_system_exit",
                [],
                ["group3.argument_parser_raise_system_exit", "START - Cookbook", "END ("],
                2,
                [],
            ),
            (
                "group3.raise_exception",
                ["RuntimeError: Something went wrong"],
                [],
                cookbook.EXCEPTION_RETCODE,
                [],
            ),
            ("group3.raise_system_exit_0", ["SystemExit(0) raised"], [], 0, []),
            ("group3.raise_system_exit_9", ["SystemExit(9) raised"], [], 9, []),
            (
                "group3.raise_system_exit_str",
                ["SystemExit('message') raised"],
                [],
                cookbook.EXCEPTION_RETCODE,
                [],
            ),
            (
                "class_api.example",
                [
                    "START - Cookbook class_api.example",
                    "END (PASS) - Cookbook class_api.example (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.multiple.CookbookA",
                [
                    "START - Cookbook class_api.multiple.CookbookA",
                    "END (PASS) - Cookbook class_api.multiple.CookbookA (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.get_runner_raise",
                [],
                ["START - Cookbook", "END ("],
                cookbook.CLASS_FAIL_INIT_RETCODE,
                [],
            ),
            (
                "class_api.lock_args",
                [
                    "START - Cookbook class_api.lock_args",
                    "END (PASS) - Cookbook class_api.lock_args (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.runtime_description",
                [
                    "START - Cookbook class_api.runtime_description Runtime description",
                    "END (PASS) - Cookbook class_api.runtime_description (exit_code=0) Runtime description",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.runtime_description_raise",
                [
                    "START - Cookbook class_api.runtime_description_raise",
                    "END (PASS) - Cookbook class_api.runtime_description_raise (exit_code=0)",
                ],
                [],
                0,
                [],
            ),
            (
                "class_api.rollback",
                [
                    "START - Cookbook class_api.rollback",
                    "run has raised",
                    "rollback called",
                    "END (FAIL) - Cookbook class_api.rollback (exit_code=99)",
                ],
                [],
                cookbook.EXCEPTION_RETCODE,
                [],
            ),
            (
                "class_api.rollback_raise",
                [
                    "START - Cookbook class_api.rollback_raise",
                    "rollback has raised",
                    "END (FAIL) - Cookbook class_api.rollback_raise (exit_code=93)",
                ],
                [],
                cookbook.ROLLBACK_FAIL_RETCODE,
                [],
            ),
            (
                "class_api.init_abort",
                [],
                [],
                0,
                [],
            ),
            (
                "class_api.init_abort",
                ["custom message"],
                [],
                0,
                ["--message", "custom message"],
            ),
        ),
    )
    def test_main_execute_cookbook(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self, tmpdir, caplog, module, messages, absent_messages, code, args
    ):
        """Calling main with the given cookbook and args should execute it."""
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with caplog.at_level(logging.INFO):
                ret = _cookbook.main([module] + args)

        assert ret == code
        for message in messages:
            assert message in caplog.text
        for message in absent_messages:
            assert message not in caplog.text

    @pytest.mark.parametrize("module", ("external_cookbook", "external_group.cookbook1"))
    def test_main_execute_external_cookbook(self, tmpdir, caplog, module):
        """Calling main with the given external cookbook and args should execute it."""
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS_MULTI,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with caplog.at_level(logging.DEBUG):
                ret = _cookbook.main([module])

        assert ret == 0
        messages = [
            f"START - Cookbook {module}",
            f"END (PASS) - Cookbook {module} (exit_code=0)",
        ]
        assert ret == 0
        for message in messages:
            assert message in caplog.text

        for line in caplog.text.splitlines():
            if "__COOKBOOK_STATS__" in line:
                break
        else:
            raise AssertionError("__COOKBOOK_STATS__ not found in logs")

        title, content = line.split()[-1].split(":")
        assert title == "__COOKBOOK_STATS__"
        stats = dict((item.split("=") for item in content.split(",")))
        assert stats["name"] == module
        assert stats["exit_code"] == "0"
        assert float(stats["duration"]) >= 0

    def test_main_execute_cookbook_invalid_args(self, tmpdir, capsys, caplog):
        """Calling a cookbook with the wrong args should let argparse print its message."""
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with caplog.at_level(logging.INFO):
                ret = _cookbook.main(["group3.argparse_ok", "--invalid"])

        assert ret == 2
        _, err = capsys.readouterr()
        assert "group3.argparse_ok: error: unrecognized arguments" in err
        for message in ("START - Cookbook", "END ("):
            assert message not in caplog.text

    def test_main_execute_dry_run(self, capsys, tmpdir):
        """Calling main() with a cookbook and dry_run mode should execute it and set the dry run mode."""
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            ret = _cookbook.main(["-d", "root"])

        assert ret == 0
        _, err = capsys.readouterr()
        assert "DRY-RUN" in err

    def test_main_execute_no_lock(self, tmpdir):
        """Calling main() with the CLI --no-lock flag mode should bypass the locking mechanism."""
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            ret = _cookbook.main(["--no-locks", "root"])

        assert ret == 0

    def test_main_list(self, tmpdir, capsys, caplog):
        """Calling main() with the -l/--list option should print the available cookbooks."""
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            with caplog.at_level(logging.INFO):
                ret = _cookbook.main(["-l"])

        out, _ = capsys.readouterr()

        assert ret == 0
        assert out == LIST_COOKBOOKS_ALL.format(external_cookbooks="")
        lines = [
            "Failed to import module cookbooks.group3.invalid_syntax: invalid syntax (invalid_syntax.py, line 6)",
            "Failed to import module cookbooks.group3.invalid_subgroup: invalid syntax (__init__.py, line 2)",
        ]
        for line in lines:
            assert line in caplog.text

    def test_cookbooks_menu_status(self, monkeypatch):
        """Calling status on a TreeItem should show the completed and total tasks."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        cookbooks = _cookbook.CookbookCollection(base_dirs=COOKBOOKS_BASE_PATHS, args=[], spicerack=self.spicerack)
        menu = cookbooks.get_item("")
        assert menu.status == "0/31"

    def test_cookbooks_menu_status_done(self, monkeypatch):
        """Calling status on a TreeItem with all tasks completed should return DONE."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        cookbooks = _cookbook.CookbookCollection(
            base_dirs=COOKBOOKS_BASE_PATHS, args=[], spicerack=self.spicerack, path_filter="group1"
        )
        menu = cookbooks.get_item("group1")
        assert menu.status == "0/1"
        item = cookbooks.get_item("group1.cookbook1")
        item.run()
        assert menu.status == "DONE"

    @mock.patch("spicerack._cookbook.CookbookItem", side_effect=_menu.MenuError("fail to init"))
    def test_cookbooks_menu_cookbook_init_fail(self, mocked_cookbook_item, caplog, monkeypatch):
        """When a CookbookItem object fail to get initialized it should catch the MenuError, log it and continue."""
        monkeypatch.syspath_prepend(COOKBOOKS_BASE_PATH)
        with caplog.at_level(logging.INFO):
            cookbooks = _cookbook.CookbookCollection(
                base_dirs=COOKBOOKS_BASE_PATHS, args=[], spicerack=self.spicerack, path_filter="cookbook"
            )
            menu = cookbooks.get_item(".".join((cookbooks.cookbooks_module_prefix, "group1")))
        assert menu is None
        assert "fail to init" in caplog.text
        assert mocked_cookbook_item.called

    def test_cookbooks_cookbook_init_fail(self):
        """When a CookbookItem object fail to get initialized it should catch the MenuError, log it and continue."""

        class NotCookbookBaseSubclass:
            """Just a class that doesn't inherit from CookbookBase."""

        with pytest.raises(_menu.MenuError, match="is not a subclass of CookbookBase"):
            _menu.CookbookItem(NotCookbookBaseSubclass, [], self.spicerack)

    @pytest.mark.parametrize(
        "tty, answer, output",
        (
            (False, "q", COOKBOOKS_MENU_NOTTY),
            (True, "q", COOKBOOKS_MENU_TTY),
            (True, KeyboardInterrupt, COOKBOOKS_MENU_TTY + "QUIT\n"),
            (True, ["", "q"], COOKBOOKS_MENU_TTY + COOKBOOKS_MENU_TTY),
            (
                True,
                ["invalid", "q"],
                COOKBOOKS_MENU_TTY + "==> Invalid input <==\n" + COOKBOOKS_MENU_TTY,
            ),
            (
                True,
                ["group1", "b", "q"],
                COOKBOOKS_MENU_TTY + COOKBOOKS_GROUP1_MENU + COOKBOOKS_MENU_TTY,
            ),
            (
                True,
                ["group2", "cookbook2", "b", "q"],
                COOKBOOKS_MENU_TTY
                + COOKBOOKS_GROUP2_MENU
                + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN
                + COOKBOOKS_MENU_TTY.replace("[0/3] group2", "[1/3] group2"),
            ),
            (
                True,
                ["group2", "cookbook2 --argument value", "b", "q"],
                COOKBOOKS_MENU_TTY
                + COOKBOOKS_GROUP2_MENU
                + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN.replace("'argument': None", "'argument': 'value'")
                + COOKBOOKS_MENU_TTY.replace("[0/3] group2", "[1/3] group2"),
            ),
            (
                True,
                ["group2 -k", "cookbook2", "b", "q"],
                COOKBOOKS_MENU_TTY
                + COOKBOOKS_GROUP2_MENU.replace("[]", "['-k']")
                + COOKBOOKS_GROUP2_COOKBOOK2_MENU_RUN.replace("[]", "['-k']").replace("'k': False", "'k': True")
                + COOKBOOKS_MENU_TTY.replace("[0/3] group2", "[1/3] group2"),
            ),
            (
                True,
                "h",
                COOKBOOKS_MENU_TTY
                + _menu.HELP_MESSAGE.format(statuses=_cookbook.CookbookItem.statuses)
                + "\n"
                + COOKBOOKS_MENU_TTY,
            ),
        ),
    )
    @mock.patch("spicerack._cookbook.sys.stdout.isatty")
    @mock.patch("builtins.input")
    def test_cookbooks_main_menu(
        self, mocked_input, mocked_tty, tty, answer, output, capsys, tmpdir
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        """Calling main() with a menu should show the menu and allow to interact with it."""
        mocked_tty.return_value = tty
        mocked_input.side_effect = answer
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            ret = _cookbook.main([])

        out, _ = capsys.readouterr()
        assert ret == 0
        assert out == output

    @mock.patch("spicerack._cookbook.sys.stdout.isatty")
    @mock.patch("builtins.input")
    def test_cookbooks_submenu(self, mocked_input, mocked_tty, capsys, tmpdir):
        """Calling main() with a menu should show the menu and allow to interact with it."""
        mocked_tty.return_value = True
        mocked_input.side_effect = ["subgroup1", "b", "q"]
        config = {
            "cookbooks_base_dirs": COOKBOOKS_BASE_PATHS,
            "logs_base_dir": tmpdir.strpath,
            "instance_params": {**SPICERACK_TEST_PARAMS},  # Make a copy
        }
        with mock.patch("spicerack._cookbook.load_yaml_config", lambda config_dir: config):
            ret = _cookbook.main(["group2"])

        out, _ = capsys.readouterr()
        assert ret == 0
        group2 = COOKBOOKS_GROUP2_MENU.replace("b - Back to parent menu", "q - Quit")
        assert out == group2 + COOKBOOKS_GROUP2_SUBGROUP1_MENU + group2
